@@ -1,6 +1,6 @@
 # vim: ft=python fileencoding=utf-8 sts=4 sw=4 et:
 
-# Copyright 2014-2020 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
+# Copyright 2014-2021 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
 #
 # This file is part of qutebrowser.
 #
@@ -15,17 +15,16 @@
 # GNU General Public License for more details.
 #
 # You should have received a copy of the GNU General Public License
-# along with qutebrowser.  If not, see <http://www.gnu.org/licenses/>.
+# along with qutebrowser.  If not, see <https://www.gnu.org/licenses/>.
 
-"""Mode manager singleton which handles the current keyboard mode."""
+"""Mode manager (per window) which handles the current keyboard mode."""
 
 import functools
+import dataclasses
 from typing import Mapping, Callable, MutableMapping, Union, Set, cast
 
-import attr
 from PyQt5.QtCore import pyqtSlot, pyqtSignal, Qt, QObject, QEvent
 from PyQt5.QtGui import QKeyEvent
-from PyQt5.QtWidgets import QApplication
 
 from qutebrowser.commands import runners
 from qutebrowser.keyinput import modeparsers, basekeyparser
@@ -33,6 +32,7 @@ from qutebrowser.config import config
 from qutebrowser.api import cmdutils
 from qutebrowser.utils import usertypes, log, objreg, utils
 from qutebrowser.browser import hints
+from qutebrowser.misc import objects
 
 INPUT_MODES = [usertypes.KeyMode.insert, usertypes.KeyMode.passthrough]
 PROMPT_MODES = [usertypes.KeyMode.prompt, usertypes.KeyMode.yesno]
@@ -40,7 +40,7 @@ PROMPT_MODES = [usertypes.KeyMode.prompt, usertypes.KeyMode.yesno]
 ParserDictType = MutableMapping[usertypes.KeyMode, basekeyparser.BaseKeyParser]
 
 
-@attr.s(frozen=True)
+@dataclasses.dataclass(frozen=True)
 class KeyEvent:
 
     """A small wrapper over a QKeyEvent storing its data.
@@ -54,8 +54,8 @@ class KeyEvent:
         text: A string (QKeyEvent::text).
     """
 
-    key = attr.ib()  # type: Qt.Key
-    text = attr.ib()  # type: str
+    key: Qt.Key
+    text: str
 
     @classmethod
     def from_event(cls, event: QKeyEvent) -> 'KeyEvent':
@@ -78,16 +78,19 @@ class UnavailableError(Exception):
 
 def init(win_id: int, parent: QObject) -> 'ModeManager':
     """Initialize the mode manager and the keyparsers for the given win_id."""
+    commandrunner = runners.CommandRunner(win_id)
+
     modeman = ModeManager(win_id, parent)
     objreg.register('mode-manager', modeman, scope='window', window=win_id)
 
-    commandrunner = runners.CommandRunner(win_id)
-
     hintmanager = hints.HintManager(win_id, parent=parent)
     objreg.register('hintmanager', hintmanager, scope='window',
                     window=win_id, command_only=True)
+    modeman.hintmanager = hintmanager
 
-    keyparsers = {
+    log_sensitive_keys = 'log-sensitive-keys' in objects.debug_flags
+
+    keyparsers: ParserDictType = {
         usertypes.KeyMode.normal:
             modeparsers.NormalKeyParser(
                 win_id=win_id,
@@ -108,7 +111,7 @@ def init(win_id: int, parent: QObject) -> 'ModeManager':
                 commandrunner=commandrunner,
                 parent=modeman,
                 passthrough=True,
-                do_log=False,
+                do_log=log_sensitive_keys,
                 supports_count=False),
 
         usertypes.KeyMode.passthrough:
@@ -118,7 +121,7 @@ def init(win_id: int, parent: QObject) -> 'ModeManager':
                 commandrunner=commandrunner,
                 parent=modeman,
                 passthrough=True,
-                do_log=False,
+                do_log=log_sensitive_keys,
                 supports_count=False),
 
         usertypes.KeyMode.command:
@@ -128,7 +131,7 @@ def init(win_id: int, parent: QObject) -> 'ModeManager':
                 commandrunner=commandrunner,
                 parent=modeman,
                 passthrough=True,
-                do_log=False,
+                do_log=log_sensitive_keys,
                 supports_count=False),
 
         usertypes.KeyMode.prompt:
@@ -138,7 +141,7 @@ def init(win_id: int, parent: QObject) -> 'ModeManager':
                 commandrunner=commandrunner,
                 parent=modeman,
                 passthrough=True,
-                do_log=False,
+                do_log=log_sensitive_keys,
                 supports_count=False),
 
         usertypes.KeyMode.yesno:
@@ -184,7 +187,7 @@ def init(win_id: int, parent: QObject) -> 'ModeManager':
                 win_id=win_id,
                 commandrunner=commandrunner,
                 parent=modeman),
-    }  # type: ParserDictType
+    }
 
     for mode, parser in keyparsers.items():
         modeman.register(mode, parser)
@@ -227,6 +230,7 @@ class ModeManager(QObject):
 
     Attributes:
         mode: The mode we're currently in.
+        hintmanager: The HintManager associated with this window.
         _win_id: The window ID of this ModeManager
         _prev_mode: Mode before a prompt popped up
         parsers: A dictionary of modes and their keyparsers.
@@ -256,10 +260,12 @@ class ModeManager(QObject):
     def __init__(self, win_id: int, parent: QObject = None) -> None:
         super().__init__(parent)
         self._win_id = win_id
-        self.parsers = {}  # type: ParserDictType
+        self.parsers: ParserDictType = {}
         self._prev_mode = usertypes.KeyMode.normal
         self.mode = usertypes.KeyMode.normal
-        self._releaseevents_to_pass = set()  # type: Set[KeyEvent]
+        self._releaseevents_to_pass: Set[KeyEvent] = set()
+        # Set after __init__
+        self.hintmanager = cast(hints.HintManager, None)
 
     def __repr__(self) -> str:
         return utils.get_repr(self, mode=self.mode)
@@ -278,8 +284,8 @@ class ModeManager(QObject):
         curmode = self.mode
         parser = self.parsers[curmode]
         if curmode != usertypes.KeyMode.insert:
-            log.modes.debug("got keypress in mode {} - delegating to "
-                            "{}".format(curmode, utils.qualname(parser)))
+            log.modes.debug("got keypress in mode {} - delegating to {}".format(
+                utils.pyenum_str(curmode), utils.qualname(parser)))
         match = parser.handle(event, dry_run=dry_run)
 
         has_modifier = event.modifiers() not in [
@@ -302,7 +308,7 @@ class ModeManager(QObject):
             self._releaseevents_to_pass.add(KeyEvent.from_event(event))
 
         if curmode != usertypes.KeyMode.insert:
-            focus_widget = QApplication.instance().focusWidget()
+            focus_widget = objects.qapp.focusWidget()
             log.modes.debug("match: {}, forward_unbound_keys: {}, "
                             "passthrough: {}, is_non_alnum: {}, dry_run: {} "
                             "--> filter: {} (focused: {!r})".format(
@@ -355,7 +361,8 @@ class ModeManager(QObject):
             return
 
         log.modes.debug("Entering mode {}{}".format(
-            mode, '' if reason is None else ' (reason: {})'.format(reason)))
+            utils.pyenum_str(mode),
+            '' if reason is None else ' (reason: {})'.format(reason)))
         if mode not in self.parsers:
             raise ValueError("No keyparser for mode {}".format(mode))
         if self.mode == mode or (self.mode in PROMPT_MODES and
@@ -381,7 +388,7 @@ class ModeManager(QObject):
         self.entered.emit(mode, self._win_id)
 
     @cmdutils.register(instance='mode-manager', scope='window')
-    def enter_mode(self, mode: str) -> None:
+    def mode_enter(self, mode: str) -> None:
         """Enter a key mode.
 
         Args:
@@ -423,7 +430,8 @@ class ModeManager(QObject):
                 raise NotInModeError("Not in mode {}!".format(mode))
 
         log.modes.debug("Leaving mode {}{}".format(
-            mode, '' if reason is None else ' (reason: {})'.format(reason)))
+            utils.pyenum_str(mode),
+            '' if reason is None else ' (reason: {})'.format(reason)))
         # leaving a mode implies clearing keychain, see
         # https://github.com/qutebrowser/qutebrowser/issues/1805
         self.clear_keychain()
@@ -433,9 +441,9 @@ class ModeManager(QObject):
             self.enter(self._prev_mode,
                        reason='restore mode before {}'.format(mode.name))
 
-    @cmdutils.register(instance='mode-manager', name='leave-mode',
+    @cmdutils.register(instance='mode-manager',
                        not_modes=[usertypes.KeyMode.normal], scope='window')
-    def leave_current_mode(self) -> None:
+    def mode_leave(self) -> None:
         """Leave the mode we're currently in."""
         if self.mode == usertypes.KeyMode.normal:
             raise ValueError("Can't leave normal mode!")
@@ -452,12 +460,12 @@ class ModeManager(QObject):
         Return:
             True if event should be filtered, False otherwise.
         """
-        handlers = {
+        handlers: Mapping[QEvent.Type, Callable[[QKeyEvent], bool]] = {
             QEvent.KeyPress: self._handle_keypress,
             QEvent.KeyRelease: self._handle_keyrelease,
             QEvent.ShortcutOverride:
                 functools.partial(self._handle_keypress, dry_run=True),
-        }  # type: Mapping[QEvent.Type, Callable[[QKeyEvent], bool]]
+        }
         handler = handlers[event.type()]
         return handler(cast(QKeyEvent, event))
 
